pub struct WhyNot(/* private fields */);Expand description
An opaque type that captures the reason why a deserialization to a
ShouldBe<T> failed.
This type is only meant to be used within the ShouldBe type.
Implementations